# flake8: noqa
from .main import glove
from .settings import BaseSettings
from .templates import FoxgloveTemplates
from .version import VERSION

__all__ = 'BaseSettings', 'glove', 'FoxgloveTemplates', 'VERSION'



import asyncio
from functools import wraps
from time import time
from typing import Optional

from starlette.templating import Jinja2Templates as _Jinja2Templates

from .main import glove

try:
    import jinja2
except ImportError:  # pragma: nocover
    jinja2 = None  # type: ignore

reload_sha = 'fbc87301a2470263e3ba45e56c7089f286a84a4e'
reload_snippet = f'<script src="https://rawcdn.githack.com/samuelcolvin/foxglove/{reload_sha}/reload.js"></script>'


class FoxgloveTemplates(_Jinja2Templates):
    def __init__(self, directory: Optional[str] = None):
        super().__init__(glove.settings.template_dir if directory is None else directory)

    def get_env(self, directory: str) -> 'jinja2.Environment':
        @jinja2.contextfunction
        def prompt_reload(context: dict) -> str:
            if context['dev_mode']:
                return jinja2.Markup(reload_snippet)
            else:
                return ''

        @jinja2.contextfunction
        def static_url(context: dict, path: str) -> str:
            request = context['request']
            url = request.url_for(context['static_route_name'], path=path)
            if context['dev_mode']:
                url += f'?t={time() * 1000:0.0f}'
            return url

        env = super().get_env(directory)
        env.globals.update(
            prompt_reload=prompt_reload,
            static_url=static_url,
            dev_mode=glove.settings.dev_mode,
            static_route_name='static',
        )
        return env

    def render(self, template_name: str):
        def view_decorator(view):
            if asyncio.iscoroutinefunction(view):

                @wraps(view)
                async def view_wrapper(request):
                    return self._return_template(request, template_name, await view(request))

            else:

                @wraps(view)
                def view_wrapper(request):
                    return self._return_template(request, template_name, view(request))

            return view_wrapper

        return view_decorator

    def _return_template(self, request, template_name: str, context: Optional[dict]):
        if context is None:
            context = {}
        context['request'] = request
        return self.TemplateResponse(template_name, context)



#!/usr/bin/env python3
import asyncio
import locale
import logging
import os
import sys
from importlib import import_module
from pathlib import Path
from typing import Callable, List

import typer
import uvloop
from pydantic.env_settings import BaseSettings as PydanticBaseSettings
from uvicorn.importer import ImportFromStringError, import_from_string
from uvicorn.main import run as uvicorn_run

from foxglove.logs import setup_logging

from .main import glove
from .settings import BaseSettings
from .version import VERSION

logger = logging.getLogger('foxglove.cli')

cli = typer.Typer()
SETTINGS_PATH: str
ROOT_PATH: Path
settings: BaseSettings


@cli.command()
def web():
    """
    Run the web server using uvicorn.
    """
    logger.info('running web server at %s...', settings.port)
    # wait_for_services(settings)
    os.environ['foxglove_settings_path'] = SETTINGS_PATH
    uvicorn_run(settings.asgi_path, host='0.0.0.0', port=settings.port, workers=settings.web_workers, log_config=None)


@cli.command()
def dev():
    """
    Run the web server using uvicorn for development
    """
    logger.info('running web server at %s in dev mode...', settings.port)
    os.environ.update(
        foxglove_dev_mode='TRUE', foxglove_settings_path=SETTINGS_PATH, foxglove_root_path=str(ROOT_PATH),
    )
    uvicorn_run(
        settings.asgi_path, host='127.0.0.1', port=settings.port, reload=True, reload_dirs=[ROOT_PATH], log_config=None,
    )


@cli.command()
def worker():
    """
    Run the worker command from settings.worker_func.
    """
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    if settings.worker_func:
        logger.info('running worker...')
        worker_func: Callable[[BaseSettings], None] = import_from_string(settings.worker_func)
        # wait_for_services(settings)
        worker_func(settings=settings)
    else:
        raise CliError("settings.worker_func not set, can't run the worker")


@cli.command()
def auto():
    """
    Run either the web server or worker depending on the environment variables: ATOOLBOX_COMMAND, DYNO and PORT.
    """
    function = _get_auto_command()
    function()


def _get_auto_command() -> Callable[[], None]:
    if command_env := os.getenv('ATOOLBOX_COMMAND'):
        logger.info('using environment variable ATOOLBOX_COMMAND=%r to infer command', command_env)
        command_env = command_env.lower()
        if command_env == 'web':
            return web
        elif command_env == 'worker':
            return worker
        elif command_env != 'auto':
            raise CliError(f'Invalid value for ATOOLBOX_COMMAND: {command_env!r}')

    if dyno_env := os.getenv('DYNO'):
        logger.info('using environment variable DYNO=%r to infer command', dyno_env)
        return web if dyno_env.lower().startswith('web') else worker
    elif (port_env := os.getenv('PORT')) and port_env.isdigit():
        logger.info('using environment variable PORT=%s to infer command as web', port_env)
        return web
    else:
        logger.info('no environment variable found to infer command, assuming worker')
        return worker


@cli.command()
def patch(patch_name: str = None, live: bool = False, extra_args: List[str] = None):
    """
    Run a patch function to update or modify the database.
    """
    logger.info('running patch...')
    from .db.patches import run_patch

    # wait_for_services(settings)
    for path in settings.patch_paths:
        import_module(path)

    return run_patch(settings, patch_name, live, extra_args)


@cli.command()
def reset_database():
    """
    Delete the main database and recreate it empty. THIS CAN BE DESTRUCTIVE!
    """
    from .db import reset_database

    logger.info('running reset_database...')
    reset_database(settings)


@cli.command()
def flush_redis():
    """
    Empty the redis cache.
    """
    from .redis import flush_redis

    logger.info('running flush_redis...')
    flush_redis(settings)


@cli.command()
def shell():
    """
    Run an interactive python shell.
    """
    from IPython import start_ipython
    from IPython.terminal.ipapp import load_default_config

    c = load_default_config()
    settings_path, settings_name = SETTINGS_PATH.split(':')
    exec_lines = [
        'import asyncio, base64, math, hashlib, json, os, pickle, re, secrets, sys, time',
        'from datetime import datetime, date, timedelta, timezone',
        'from pathlib import Path',
        'from pprint import pprint as pp',
        '',
        'sys.path.append(os.getcwd())',
        f'ROOT_PATH = Path("{ROOT_PATH}")',
        'sys.path.append(str(ROOT_PATH))',
        'os.chdir(str(ROOT_PATH))',
        '',
        f'from {settings_path} import {settings_name}',
        'settings = Settings()',
    ]
    exec_lines += ['print("\\n    Python {v.major}.{v.minor}.{v.micro}\\n".format(v=sys.version_info))'] + [
        f"print('    {line}')" for line in exec_lines
    ]

    c.TerminalIPythonApp.display_banner = False
    c.TerminalInteractiveShell.confirm_exit = False
    c.InteractiveShellApp.exec_lines = exec_lines

    start_ipython(argv=(), config=c)


@cli.callback(help=f'foxglove command line interface v{VERSION}')
def callback(
    settings_path: str = typer.Option(
        os.getenv('ATOOLBOX_SETTINGS', 'settings:Settings'),
        '-s',
        help=(
            'settings path (dotted, relative to the root directory), defaults to to the environment variable '
            '"ATOOLBOX_SETTINGS" or "settings.Settings"'
        ),
    ),
    root: str = typer.Option(
        os.getenv('ATOOLBOX_ROOT_DIR', '.'),
        '-r',
        help='root directory to run command from, defaults to to the environment variable "ATOOLBOX_ROOT_DIR" or "."',
    ),
) -> None:
    # ugly work around, is there another way? https://github.com/tiangolo/typer/issues/55
    if {'--help', '--version'} & set(sys.argv):
        return
    global ROOT_PATH, SETTINGS_PATH, settings
    setup_logging()

    sys.path.append(os.getcwd())
    ROOT_PATH = Path(root).resolve()
    sys.path.append(str(ROOT_PATH))
    os.chdir(str(ROOT_PATH))

    if ':' not in settings_path:
        settings_path += ':Settings'

    SETTINGS_PATH = settings_path
    try:
        settings_cls = import_from_string(settings_path)
    except ImportFromStringError as exc:
        raise CliError(f'unable to import "{settings_path}", {exc.__class__.__name__}: {exc}')

    if not isinstance(settings_cls, type) or not issubclass(settings_cls, PydanticBaseSettings):
        raise CliError(f'settings "{settings_cls}" (from "{settings_path}"), is not a valid Settings class')

    settings = settings_cls()
    settings_locale = getattr(settings, 'locale', None)
    if settings_locale:
        locale.setlocale(locale.LC_ALL, settings_locale)

    glove.settings = settings


class CliError(typer.Exit):
    def __init__(self, msg=None, code: int = 1):
        print(msg, file=sys.stdout)
        super().__init__(code)


if __name__ == '__main__':  # pragma: no cover
    cli()



import asyncio

import httpx
from buildpg.asyncpg import BuildPgPool

from .db import create_pg_pool
from .settings import BaseSettings

__all__ = ('glove',)


class Glove:
    settings: BaseSettings
    pg: BuildPgPool
    http: httpx.AsyncClient

    async def startup(self):
        self.pg = await create_pg_pool(self.settings)
        self.http = httpx.AsyncClient(timeout=self.settings.http_client_timeout)

    async def shutdown(self):
        await asyncio.gather(self.pg.close(), self.http.aclose())


glove = Glove()



from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import setup

description = 'Tools for Starlette'
THIS_DIR = Path(__file__).resolve().parent
try:
    long_description = THIS_DIR.joinpath('README.md').read_text()
except FileNotFoundError:
    long_description = description

# avoid loading the package before requirements are installed:
version = SourceFileLoader('version', 'foxglove/version.py').load_module()

setup(
    name='foxglove-web',
    version=str(version.VERSION),
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX :: Linux',
        'Environment :: MacOS X',
        'Topic :: Internet',
    ],
    author='Samuel Colvin',
    author_email='s@muelcolvin.com',
    url='https://github.com/samuelcolvin/foxglove',
    license='MIT',
    packages=['foxglove', 'foxglove.db'],
    entry_points="""
        [console_scripts]
        foxglove=foxglove.__main__:cli
    """,
    python_requires='>=3.8',
    zip_safe=True,
    install_requires=[
        'arq>=0.18',
        'asyncpg>=0.20.1',
        'starlette>=0.13.2',
        'buildpg>=0.2.1',
        'httpx>=0.11.1',
        'pydantic>=1.4',
        'sentry-sdk>=0.14',
        'typer>=0.1.0',
        'uvicorn>=0.11.3',
    ],
    extras_require={'extra': ['ipython>=7.7.0']},
)



import os

from starlette.applications import Starlette
from starlette.middleware import Middleware
from uvicorn.importer import import_from_string

from .db import PgMiddleware
from .devtools import reload_endpoint
from .main import glove
from .settings import BaseSettings

Settings = import_from_string(os.environ['foxglove_settings_path'])
settings: BaseSettings = Settings()

glove.settings = settings
middleware = []
if settings.pg_dsn:
    middleware += [Middleware(PgMiddleware)]

routes = settings.get_routes()
if settings.dev_mode:
    foxglove_root_path = os.environ.get('foxglove_root_path')

    if foxglove_root_path:
        routes += reload_endpoint(foxglove_root_path)
    else:
        raise RuntimeError('dev_mode enabled but "foxglove_root_path" not found, can\'t add the reload endpoint')


app = Starlette(
    debug=settings.dev_mode,
    middleware=middleware,
    routes=settings.get_routes(),
    on_startup=[glove.startup],
    on_shutdown=[glove.shutdown],
)



from pathlib import Path
from typing import List, Optional, Pattern
from urllib.parse import urlparse

from pydantic import BaseSettings as PydanticBaseSettings, validator
from starlette.routing import Route
from uvicorn.importer import import_from_string

try:
    from arq.connections import RedisSettings

    redis_settings_default = 'redis://localhost:6379'
except ImportError:
    redis_settings_default = None

    class RedisSettings:
        """
        Mock arq.RedisSettings to satisfy pydantic if arq isn't installed
        """

        def __init__(self, *args, **kwargs):
            raise RuntimeError('arq not installed')


try:
    from buildpg import asyncpg  # noqa
except ImportError:
    pg_dsn_default = None
else:
    pg_dsn_default = 'postgres://postgres@localhost:5432/app'


class BaseSettings(PydanticBaseSettings):
    dev_mode: bool = False
    asgi_path: str = 'foxglove.asgi:app'
    routes: Optional[str] = None
    web_workers: Optional[int] = None

    worker_func: Optional[str] = None

    patch_paths: List[str] = []

    sql_path: Path = 'models.sql'
    template_dir: Optional[Path] = 'templates'
    pg_dsn: Optional[str] = pg_dsn_default
    # eg. the db already exists on heroku and never has to be created
    pg_db_exists = False
    pg_pool_min_size: int = 10
    pg_pool_max_size: int = 10

    redis_settings: Optional[RedisSettings] = redis_settings_default
    port: int = 8000

    locale: Optional[str] = None

    http_client_timeout = 10

    csrf_ignore_paths: List[Pattern] = []
    csrf_upload_paths: List[Pattern] = []
    csrf_cross_origin_paths: List[Pattern] = []
    cross_origin_origins: List[Pattern] = []

    grecaptcha_url = 'https://www.google.com/recaptcha/api/siteverify'

    # this is the test key from https://developers.google.com/recaptcha/docs/faq, or
    # https://developers.google.com/recaptcha/docs/faq#id-like-to-run-automated-tests-with-recaptcha-what-should-i-do
    # you'll need to change it for production
    grecaptcha_secret = '6LeIxAcTAAAAAGG-vFI1TnRWxMZNFuojJ4WifJWe'

    @property
    def sql(self):
        return self.sql_path.read_text()

    def get_routes(self) -> List[Route]:
        routes = self.routes or f'{self.__module__}:routes'
        return import_from_string(routes)

    @property
    def _pg_dsn_parsed(self):
        return urlparse(self.pg_dsn)

    @property
    def pg_name(self):
        return self._pg_dsn_parsed.path.lstrip('/')

    @property
    def pg_host(self):
        return self._pg_dsn_parsed.hostname

    @property
    def pg_port(self):
        return self._pg_dsn_parsed.port

    @validator('redis_settings', always=True, pre=True)
    def parse_redis_settings(cls, v):
        if v is None:
            return

        if RedisSettings.__module__ != 'arq.connections':
            raise RuntimeError(f'arq must be installed to use redis, redis_settings set to {v!r}')
        conf = urlparse(v)
        return RedisSettings(
            host=conf.hostname, port=conf.port, password=conf.password, database=int((conf.path or '0').strip('/'))
        )

    class Config:
        fields = {
            'pg_dsn': {'env': 'DATABASE_URL'},
            'redis_settings': {'env': ['REDISCLOUD_URL', 'REDIS_URL']},
            'dev_mode': {'env': ['foxglove_dev_mode']},
        }
